# This file contains general functions for various evolutionary algorithms

import numpy as np
import copy
from math import *
import random
import csv
from numpy import linalg as LA
from random import shuffle

EPS = 1.0e-14


def check_dominate(v1, v2):
    s = (v1 < v2).sum()
    if s == len(v1):
        return 1
    elif s == 0:
        return -1
    else:
        return 0


def check_dominate_(x1, x2):
    flag1 = 0
    flag2 = 0
    for i in range(0, len(x1)):
        if (x1[i] < x2[i]):
            flag1 = 1
        if (x1[i] > x2[i]):
            flag2 = 1
    if (flag1 == 1 and flag2 == 0):
        return 1  # x1 dominates x2 (for min)
    elif (flag2 == 1 and flag1 == 0):
        return -1  # x2 dominates x1	(for min)
    else:
        return 0  # non-dominated


def nd(archive):
    nd_data = []
    n = []
    for i in range(0, len(archive)):
        n.append(0)
        for j in range(0, len(archive)):
            if j != i:
                if (check_dominate(archive[i][1], archive[j][1]) == -1):
                    n[i] = n[i] + 1
        if n[i] == 0:
            nd_data.append(archive[i])
    return nd_data


def purity(data_arr):
    for i in range(0, len(data_arr)):
        for j in range(0, len(data_arr[i])):
            data_arr[i][j] = np.append(data_arr[i][j], i)
            # data_arr[i][j].append(i)
    t_data = []
    for i in range(0, len(data_arr)):
        t_data = t_data + data_arr[i]
    nd_data = nd(t_data)
    r_cnt = [0] * len(data_arr)
    for i in range(0, len(nd_data)):
        k = int(nd_data[i][len(nd_data[i]) - 1])
        r_cnt[k] = r_cnt[k] + 1.0
    for i in range(0, len(r_cnt)):
        r_cnt[i] = r_cnt[i] / len(data_arr[i])
    return r_cnt


def d2(data, point):
    size = len(data)
    n1 = LA.norm(point)
    point = (np.dot(data, point) / (n1 ** 2)) * point
    return LA.norm(data - point)


def IGD(ref_points, archive):
    d = 0.0
    for i in range(0, len(ref_points)):
        min_d = LA.norm(ref_points[i] - archive[0][1])
        for j in range(1, len(archive)):
            dd = LA.norm(ref_points[i] - archive[j][1])
            if dd < min_d:
                min_d = dd
        d = d + min_d
    d = d / len(ref_points)
    return d


def real_mutate(v, b, min_x, max_x):
    v_new = copy.deepcopy(v)
    r = random.randint(0, len(v) - 1)
    y = v_new[r]
    d_rnd = random.random() - 0.5
    d_rnd_lap = 0.0

    if (d_rnd < 0):
        d_rnd_lap = b * log(1.0 - 2.0 * abs(d_rnd))
    else:
        d_rnd_lap = -1.0 * b * log(1.0 - 2.0 * abs(d_rnd))
    y = y + d_rnd_lap
    i_count = 0
    while (((y < min_x[r]) or (y > max_x[r])) and (i_count < 20)):
        y = v_new[r]
        d_rnd = random.random() - 0.5
        d_rnd_lap = 0.0

        if (d_rnd < 0.0):
            d_rnd_lap = b * log(1.0 - (2.0 * abs(d_rnd)))
        else:
            d_rnd_lap = -1.0 * b * log(1.0 - (2.0 * abs(d_rnd)))
        y = y + d_rnd_lap
        i_count = i_count + 1
    v_new[r] = y
    if (i_count == 20):
        if (v_new[r] < min_x[r]):
            v_new[r] = min_x[r]
        elif (v_new[r] > max_x[r]):
            v_new[r] = max_x[r]
    return v_new


def polynomial_mutate(v, mut_prob, eta_m, min_x, max_x):
    v_new = copy.deepcopy(v)
    for i in range(0, len(v)):
        r = random.random()
        if (r < mut_prob):
            y = v_new[i]
            yl = min_x[i]
            yu = max_x[i]

            delta1 = (y - yl) / (yu - yl)
            delta2 = (yu - y) / (yu - yl)

            mut_pow = 1.0 / (eta_m + 1.0)
            deltaq = 0.0

            rnd = random.random()

            if (rnd <= 0.5):
                xy = 1.0 - delta1
                val = (2.0 * rnd) + ((1.0 - (2.0 * rnd)) * (xy ** (eta_m + 1.0)))
                deltaq = (val ** mut_pow) - 1.0
            else:
                xy = 1.0 - delta2
                val = (2.0 * (1.0 - rnd)) + (2.0 * (rnd - 0.5) * (xy ** (eta_m + 1.0)))
                deltaq = 1.0 - (val ** mut_pow)

            y = y + deltaq * (yu - yl)
            y = min(yu, max(yl, y))
            v_new[i] = y
    return v_new


def SBX(parent1, parent2, eta_c, cross_prob, min_x, max_x):
    child1 = copy.deepcopy(parent1)
    child2 = copy.deepcopy(parent2)
    r = random.random()
    if (r < cross_prob):
        for i in range(0, len(parent1)):
            r = random.random()
            if (r <= 0.5):
                if (abs(parent1[i] - parent2[i]) > EPS):
                    y1 = min(parent1[i], parent2[i])
                    y2 = max(parent1[i], parent2[i])
                    yl = min_x[i]
                    yu = max_x[i]
                    r = random.random()
                    beta = 1.0 + (2.0 * (y1 - yl) / (y2 - y1))
                    alpha = 2.0 - (beta ** (-1.0 * (1.0 + eta_c)))
                    betaq = 0.0
                    if (r <= (1.0 / alpha)):
                        betaq = (r * alpha) ** (1.0 / (eta_c + 1.0))
                    else:
                        betaq = (1.0 / (2.0 - (r * alpha))) ** (1.0 / (eta_c + 1.0))
                    child1[i] = 0.5 * (y1 + y2 - (betaq * (y2 - y1)))

                    beta = 1.0 + (2.0 * (yu - y2) / (y2 - y1))
                    alpha = 2.0 - (beta ** (-1.0 * (1.0 + eta_c)))
                    if (r <= (1.0 / alpha)):
                        betaq = (r * alpha) ** (1.0 / (eta_c + 1.0))
                    else:
                        betaq = (1.0 / (2.0 - (r * alpha))) ** (1.0 / (eta_c + 1.0))
                    child2[i] = 0.5 * ((y1 + y2) + (betaq * (y2 - y1)))

                    child1[i] = min(yu, max(yl, child1[i]))
                    child2[i] = min(yu, max(yl, child2[i]))

                    r = random.random()
                    if (r <= 0.5):
                        sw = child1[i]
                        child1[i] = child2[i]
                        child2[i] = sw

    return child1, child2


def SBX_mutate(arch, p1, eta_c, cross_prob, min_x, max_x):
    size = len(p1)
    size2 = len(arch)
    r = random.randint(0, size2 - 1)
    p2 = copy.deepcopy(arch[r][0])
    c1, c2 = SBX(p1, p2, eta_c, cross_prob, min_x, max_x)
    v_new = []
    rnd = random.random()
    if (rnd > 0.5):
        v_new = c1
    else:
        v_new = c2
    return v_new


def SBX_mutate_2(history, nbr, p1, eta_c, cross_prob, min_x, max_x):
    size = len(p1)
    a = random.sample(range(0, len(nbr)), 1)
    r1 = random.randint(0, len(history[nbr[a[0]]]) - 1)
    p2 = copy.deepcopy(history[nbr[a[0]]][r1][0])

    c1, c2 = SBX(p1, p2, eta_c, cross_prob, min_x, max_x)
    v_new = []
    rnd = random.random()
    if (rnd > 0.5):
        v_new = c1
    else:
        v_new = c2
    return v_new


def diff_mutate(archive, v, F, CR, min_x, max_x):
    size = len(v)
    a = random.sample(range(0, len(archive)), 3)
    p1 = copy.deepcopy(archive[a[0]][0])
    p2 = copy.deepcopy(archive[a[1]][0])
    p3 = copy.deepcopy(archive[a[2]][0])

    k_rand = random.randint(0, size - 1)
    u = copy.deepcopy(v)
    for k in range(0, size):
        r = random.random()
        if ((r < CR) or (k == k_rand)):
            u[k] = min(max_x[k], max(p1[k] + F * (p2[k] - p3[k]), min_x[k]))
    return u


def diff_mutate_2(history, nbr, v, F, CR, min_x, max_x):
    size = len(v)
    a = random.sample(range(0, len(nbr)), 3)
    r1 = random.randint(0, len(history[nbr[a[0]]]) - 1)
    p1 = copy.deepcopy(history[nbr[a[0]]][r1][0])
    r2 = random.randint(0, len(history[nbr[a[1]]]) - 1)
    p2 = copy.deepcopy(history[nbr[a[1]]][r2][0])
    r3 = random.randint(0, len(history[nbr[a[2]]]) - 1)
    p3 = copy.deepcopy(history[nbr[a[2]]][r3][0])

    k_rand = random.randint(0, size - 1)
    u = copy.deepcopy(v)
    for k in range(0, size):
        r = random.random()
        if ((r < CR) or (k == k_rand)):
            u[k] = min(max_x[k], max(p1[k] + F * (p2[k] - p3[k]), min_x[k]))
    return u


def crowd_sort(data, mx, mn):
    inf = 0
    for i in range(0, len(data)):
        data[i][1] = np.append(data[i][1], 0)
    d = len(data[0][1]) - 1
    for j in range(0, d - 1):
        data = sorted(data, key=lambda x: x[1][j])  #############
        data[0][1][d] = -1
        data[len(data) - 1][1][d] = -1  # assign largest distance to boundary solutions
        for i in range(1, len(data) - 1):
            if data[i][1][d] != -1:
                data[i][1][d] = data[i][1][d] + ((data[i + 1][1][j] - data[i - 1][1][j]) / (
                    mx[j] - mn[j]))  # distance = half of perimeter of bounding rectangle
                if data[i][1][d] > inf:  # for finding max so far
                    inf = data[i][1][d]
    for i in range(0, len(data)):
        if data[i][1][d] == -1:
            data[i][1][d] = inf + 1  # assigning the infinity as the number greater than the max
    data = sorted(data, key=lambda x: x[1][d], reverse=True)  # sorting in reverse
    for i in range(0, len(data)):
        data[i][1] = np.delete(data[i][1], len(data[i][1]) - 1)
    return data


def nds(data, population_size):
    n = []
    S = []
    P = [[]]
    P_d = [[]]
    for i in range(0, len(data)):
        n.append(0)
        S_ = []
        for j in range(0, len(data)):
            if (j != i):
                d = check_dominate_(data[i][1], data[j][1])
                if (d == 1):
                    S_.append(j)
                if (d == -1):
                    n[i] = n[i] + 1
        S.append(S_)
        if (n[i] == 0):
            P[0].append(i)
            P_d[0].append(data[i])
    k = 0
    l = len(P[k])
    while (l < population_size):
        Q = []
        Q_d = []
        for p in P[k]:
            for q in S[p]:
                n[q] = n[q] - 1
                if (n[q] == 0):
                    Q.append(q)
                    Q_d.append(data[q])
        k = k + 1
        l = l + len(Q)
        P.append(Q)
        P_d.append(Q_d)

    return P_d, l


def nds_th(arch, norm, archive_ptr, theta, l_ref):
    ref_sets = []
    l = 0
    for i in range(0, l_ref):
        rf = []
        for j in range(0, len(archive_ptr)):
            if (archive_ptr[j][0] == i):
                fitness = sqrt((LA.norm(norm[j]) ** 2) - (archive_ptr[j][1] ** 2)) + (theta * archive_ptr[j][1])
                mem = [arch[j], fitness]
                rf.append(mem)
        rf.sort(key=lambda x: x[1])
        if len(rf) > l:
            l = len(rf)
        ref_sets.append(rf)

    final = []
    for i in range(0, l):
        rank = []
        for j in range(0, len(ref_sets)):
            if len(ref_sets[j]) > i:
                rank.append(ref_sets[j][i])
        final.append(rank)
    for i in range(0, len(final)):
        for j in range(0, len(final[i])):
            final[i][j] = final[i][j][0]
    return final


def niching(K, pop_init, next_front, assoc_arr, ref_pts):
    k = 0
    niche_cnt = []
    l = len(pop_init)
    for i in range(0, len(ref_pts)):
        cnt = 0
        for j in range(0, l):
            if assoc_arr[j][0] == i:
                cnt = cnt + 1
        niche_cnt.append(cnt)
        cnt = 0
    niche_cnt = np.asarray(niche_cnt)
    while k < K:
        j_min = np.argmin(niche_cnt)
        Ij = []
        for i in range(l, len(assoc_arr)):
            if assoc_arr[i][0] == j_min:
                Ij.append(i)
        if len(Ij) == 0:
            niche_cnt[j_min] = len(assoc_arr) + 1
        else:
            if niche_cnt[j_min] == 0:
                t = []
                for i in range(0, len(Ij)):
                    a = []
                    a.append(Ij[i])
                    a.append(assoc_arr[Ij[i]][1])
                    t.append(a)
                t = sorted(t, key=lambda x: x[1])
                chosen_one = t[0][0]
                pop_init.append(next_front.pop(chosen_one - l))
                assoc_arr.insert(l, assoc_arr.pop(chosen_one))
            else:
                chosen_one = Ij[random.randint(0, len(Ij) - 1)]
                pop_init.append(next_front.pop(chosen_one - l))
                assoc_arr.insert(l, assoc_arr.pop(chosen_one))
            niche_cnt[j_min] = niche_cnt[j_min] + 1
            k = k + 1
            l = l + 1
    return pop_init


def form_mating_pool(par):
    shuffle(par)
    return par


def tournament_selection(par):
    par1 = list(range(len(par)))
    par2 = list(range(len(par)))
    shuffle(par1)
    shuffle(par2)
    mating_pool1 = []
    mating_pool2 = []
    for i in range(0, len(par1), 2):
        if par1[i] < par1[i + 1]:
            mating_pool1.append(par[par1[i]])
        else:
            mating_pool1.append(par[par1[i + 1]])
    for i in range(0, len(par2), 2):
        if par2[i] < par2[i + 1]:
            mating_pool2.append(par[par2[i]])
        else:
            mating_pool2.append(par[par2[i + 1]])

    return mating_pool1, mating_pool2


def ASF(objs, weight):
    w = weight[0]
    if w == 0:
        w = 0.00001
    max_ratio = objs[0] / w
    for i in range(1, len(objs)):
        w = weight[i]
        if w == 0:
            w = 0.00001
        max_ratio = max(max_ratio, objs[i] / w)
    return max_ratio


def FindMaxObjectives(archive):
    dim = len(archive[0][1])
    a_size = len(archive)

    max_point = copy.deepcopy(archive[0][1])

    for i in range(0, dim):
        for j in range(1, a_size):
            val = archive[j][1][i]
            if (val > max_point[i]):
                max_point[i] = val
    return max_point


def get_extreme_pts(archive):
    ideal = copy.deepcopy(archive[0][1])
    dim = len(archive[0][1])
    a_size = len(archive)

    norm = []
    for i in range(0, dim):
        for j in range(1, a_size):
            val = archive[j][1][i]
            if val < ideal[i]:
                ideal[i] = val
    for i in range(0, a_size):
        norm.append(archive[i][1] - ideal)

    extreme_points = []
    for f in range(0, dim):
        w = np.asarray([0.000001] * dim)
        w[f] = 1.0
        min_ASF = ASF(norm[0], w)
        min_indv = 0
        for i in range(1, a_size):
            asf = ASF(norm[i], w)
            if (asf < min_ASF):
                min_ASF = asf
                min_indv = i
        extreme_points.append(min_indv)

    duplicate = False
    i = 0
    while ((i < len(extreme_points)) and (duplicate == False)):
        j = i + 1
        while ((j < len(extreme_points)) and (duplicate == False)):
            duplicate = (extreme_points[i] == extreme_points[j])
            j += 1
        i += 1

    intercepts = np.asarray([0.0] * dim)
    negative_intercept = False
    if (duplicate == False):
        b = np.asarray([1.0] * dim)
        A = []
        for p in range(0, dim):
            A.append(copy.deepcopy(norm[extreme_points[p]]))
        A = np.asarray(A)
        x = np.dot(LA.inv(A), b)
        for f in range(0, dim):
            intercepts[f] = 1.0 / x[f]
            if (x[f] < 0):
                negative_intercept = True
                break

    if (duplicate or negative_intercept):
        max_objs = FindMaxObjectives(archive)
        for f in range(0, dim):
            intercepts[f] = max_objs[f]

    return intercepts, ideal


def normalize(archive, ideal, intercepts):
    norm = []
    for i in range(0, len(archive)):
        norm.append(archive[i][1] - ideal)

    for i in range(0, len(norm)):
        for f in range(0, len(archive[0][1])):
            if (abs(intercepts[f]) > 10e-10):
                norm[i][f] = norm[i][f] / intercepts[f]
            else:
                norm[i][f] = norm[i][f] / 10e-10
    return norm


def associate_pt(pt, ref_pts):
    data_pair = []
    data_pair.append(0)
    data_pair.append(d2(pt, ref_pts[0]))
    for i in range(1, len(ref_pts)):
        d_val = d2(pt, ref_pts[i])
        if d_val < data_pair[1]:
            data_pair[1] = d_val
            data_pair[0] = i
    return data_pair


def associate(norm, ref_pts):
    pop_arr = np.zeros(len(ref_pts))
    archive_pointer = []
    for i in range(0, len(norm)):
        data_pair = associate_pt(norm[i], ref_pts)
        archive_pointer.append(data_pair)
        pop_arr[int(data_pair[0])] += 1
    return pop_arr, archive_pointer


def readfile(dim_out):
    fname = "DTLZ/DTLZ1(" + str(dim_out) + ").csv"
    ref_points = []
    with open(fname) as csvfile:
        csvreader = csv.reader(csvfile, delimiter=" ")
        for line in csvreader:
            ref_points.append(line)

    for i in range(0, len(ref_points)):
        x = []
        for j in range(0, len(ref_points[i]) - 1):
            x.append(float(ref_points[i][j]))
        ref_points[i] = np.asarray(x)
    return ref_points


def read_ref_pts(func, dim_out):
    ref_points = []
    if func == "DTLZ1":
        fname = "DTLZ/DTLZ1(" + str(dim_out) + ").csv"
        with open(fname) as csvfile:
            csvreader = csv.reader(csvfile, delimiter=" ")
            for line in csvreader:
                ref_points.append(line)

        for i in range(0, len(ref_points)):
            x = []
            for j in range(0, len(ref_points[i]) - 1):
                x.append(float(ref_points[i][j]))
            ref_points[i] = np.asarray(x)
    else:
        fname = "DTLZ/DTLZ2(" + str(dim_out) + ").csv"
        with open(fname) as csvfile:
            csvreader = csv.reader(csvfile, delimiter=" ")
            for line in csvreader:
                ref_points.append(line)

        for i in range(0, len(ref_points)):
            x = []
            for j in range(0, len(ref_points[i]) - 1):
                x.append(float(ref_points[i][j]))
            ref_points[i] = np.asarray(x)
    return ref_points


def form_ref_table(ref_pts, nbr_size):
    table = []
    for i in range(0, len(ref_pts)):
        dist = []
        ind = []
        for j in range(0, len(ref_pts)):
            d = LA.norm(ref_pts[i] - ref_pts[j])
            dist.append(d)
            ind.append(j)
        ind = [x for (y, x) in sorted(zip(dist, ind))]
        table.append(ind[1:nbr_size])
    return table


class form_ref_pts(object):
    def __init__(self, m, divisions):
        self.M = m - 1
        self.div = divisions
        self.points = []

    def recursive(self, arr, d, l):
        arr_c = copy.deepcopy(arr)
        if d == self.M - 1:
            self.points.append(arr_c)
        else:
            for i in range(0, l):
                node_val = float(i) / float(self.div)
                arr_next = copy.deepcopy(arr_c)
                arr_next.append(node_val)
                self.recursive(arr_next, d + 1, l - i)

    def form(self):
        layer = []
        for i in range(0, self.div + 1):
            layer.append(float(i) / float(self.div))
        for i in range(0, len(layer)):
            l1 = []
            l1.append(layer[i])
            self.recursive(l1, 0, len(layer) - i)
        for i in range(0, len(self.points)):
            s = sum(self.points[i])
            self.points[i].append(1.0 - s)
        self.points = np.asarray(self.points)
