# NAEMO: Neighborhood-sensitive Archived Evolutionary Many-objective Optimization Algorithm
# by Raunak Sengupta, Monalisa Pal, Sriparna Saha, Sanghamitra Bandyopadhyay (2018)

# This is the main program which executes the procedure from a code named 'naemo_algo'
# based on the settings provided by various parameters

from naemo_algo import *

########################## EXAMPLE - 1 ##########################
# Parameter Settings
problem = 'DTLZ1'  # problem
n_obj = 3  # number of objectives
n_var = 7  # number of decision variables

maxgen = 400  # maximum generation (problem-specific)
L_hard = 91  # lower limit of the size of global archive (equal to number of reference line), specific to n_obj
L_soft = 100  # upper limit of the size of global archive, specific to n_obj
div = 12  # number of division for defining reference lines (one layered)
mut_prob = 0.75  # mutation probability for mutation switching scheme
flag1 = False  # problem specific parameter (for mutation switching scheme)
flag2 = True  # problem specific parameter (for mutation switching scheme)

# Create an instance of class NAEMO and perform optimization
optimizer = NAEMO()
optimizer.define_exp(maxgen, L_hard, L_soft, div, mut_prob, problem, n_var, n_obj, flag1, flag2)
optimizer.optimize()

# Performance evaluation in terms of IGD
igd = optimizer.evaluate_opt()
print('IGD for DTLZ1(3): ' + str(igd))

########################## EXAMPLE - 2 ##########################
# Parameter Settings
problem = 'DTLZ2'  # problem
n_obj = 10  # number of objectives
n_var = 19  # number of decision variables

maxgen = 750  # maximum generation (problem-specific)
L_hard = 275  # lower limit of the size of global archive (equal to number of reference line), specific to n_obj
L_soft = 280  # upper limit of the size of global archive, specific to n_obj
div = 3  # number of division for defining reference lines (one layered)
mut_prob = 0.75  # mutation probability for mutation switching scheme
flag1 = False  # problem specific parameter (for mutation switching scheme)
flag2 = False  # problem specific parameter (for mutation switching scheme)

# Create an instance of class NAEMO and perform optimization
optimizer = NAEMO()
optimizer.define_exp(maxgen, L_hard, L_soft, div, mut_prob, problem, n_var, n_obj, flag1, flag2)
optimizer.optimize()

# Performance evaluation in terms of IGD
igd = optimizer.evaluate_opt()
print('IGD for DTLZ2(10): ' + str(igd))
